#ifdef AIX
#define BSD_INCLUDES
#endif

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>

#ifndef AIX
#include <sys/timeb.h>
#endif

#include <math.h>

#if defined(M_XENIX) && !defined(M_UNIX)
# include <sys/types.tcp.h>
# include <sys/errno.tcp.h>
#endif

#include <netinet/in.h>
#include <arpa/telnet.h>
#include <errno.h>
#include "config.h"
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "realms.h"
#include "wiz_list.h"
#include "error.h"
#include "o_hash.h"
#include "comm.h"

#define MAX_RESET 2700

/*
 * The 'current_time' is updated at every call cycle with time(0).
 */
long current_time,hb_time;
int num_resets_in_list;

struct realms *load_realm;

char *string_copy PROT((char *));

extern int check_last_cmd();
extern char *get_last_cmd();
extern void set_last_cmd PROT((char *));
extern struct object_status *ob_status[];
extern int num_player;
struct link_list *current_heart_beat;
struct link_list *next_reset;

extern long time();

#if !defined(linux)
extern void ftime();
#endif

extern void remove_destructed_objects(),
	    slow_shut_down(),
	    add_message();

extern unsigned sleep();

void call_heart_beat(), calls_heart_beats();
void clear_calls(),clear_heart_beats();
void load_first_objects(), prepare_ipc(),
    shutdowngame(), ed_cmd PROT((char *)),
    print_prompt(), move_current_reset PROT((int)), call_out(),
    destruct2 PROT((struct object *));

extern player_parser PROT((char *)),
    call_function_interactive PROT((struct interactive *, char *)),
    swap PROT((struct object *));

VALUE *query_load_av();

extern long start_time;
extern char new_error;
int num_resets_called=0;

void logon(ob,fun,arg)
    struct object *ob;
    char *fun;
    struct value *arg;
{
	extern int eval_cost;
    struct object_status *save_ob_status[2];
	jmp_buf erc;
	struct jmp_globals jmp;

    /*
     * Set current object to ob, so that the static function "logon"
     * can be called.
     */
    save_ob_status[0]=ob_status[0];
    save_ob_status[1]=ob_status[1];
    ob_status[1]=ob_status[0];
    ob_status[0]=alloc_ob_status();
    ob_status[0]->ob = ob;
	set_jump_state(erc,&jmp);
	if(setjmp(error_recovery_context)) 
	  {  add_message("logon: Error cannot logon.\n"); 
		 clear_jump_state(&jmp);
		 destruct_object(ob,0);
	  }
	else if(!apply(0,fun,ob,arg,0))
      {
		debug_message(ob->realm,"logon: Could not find the function '%s' in the object %s\n",fun,ob->name);
	    add_message("logon: Could not find the function '%s' in the object %s\n",fun, ob->name);
	    destruct_object(ob,0);
      }
	reset_jump_state(erc);
    free_ob_status(ob_status[0]);
    ob_status[0]=save_ob_status[0];
    ob_status[1]=save_ob_status[1];
}

/*
 * Take a player command and parse it.
 * The command can also come from a NPC.
 * Beware that 'str' can be modified and extended !
 */
int parse_command(str, ob)
    char *str;
    struct object *ob;
{
    struct object *save = command_giver;
    int res=0;

    if(str=='\0' || strlen(str)==0) return 0;
    command_giver = ob;
    res = player_parser(str);
    command_giver = save;
    return res;
}

/*
 * This is the backend. We will stay here for ever (almost).
 */

#ifndef AIX
struct timeb wait_time;
#endif

int eval_cost=0;

unsigned char more_cmds=0;

void backend()
{
    extern char game_is_being_shut_down,slow_shut_down_to_do;

    (void)signal(SIGHUP, SIG_IGN);
    (void)signal(SIGTERM, shutdowngame);
    (void)signal(SIGKILL, shutdowngame);
    (void)signal(SIGILL , shutdowngame);
    (void)signal(SIGIOT , shutdowngame);
#ifndef AIX
    (void)signal(SIGABRT, shutdowngame);
#endif

#ifndef sun
    (void)signal(SIGSEGV, shutdowngame);
#endif

     ob_status[0]=alloc_ob_status();

     move_current_reset(0);  /* This is to initialize the delay */

#if 0
{
 int load_time;


    load_time=time(0l);
    printf("Total Load Time: %d\n",(time(0l)-load_time));

}
#endif
    load_first_objects();
    load_realm=0;
    new_error=0;
    debug_message(sys_realm,"Calling initial call outs...");
    call_out();

    debug_message(sys_realm,"Done.\nSetting up ipc.\n");
    prepare_ipc();

    debug_message(sys_realm,"Mud is now running.\n");

    hb_time=time(0l);

#ifndef AIX
  ftime(&wait_time);
#endif

  while(1)
   {
	extern int num_error;
	extern void check_for_connects(),purge_dead_sockets(),
		    get_messages();
	void update_load_av PROT((void));

	extern struct interactive *all_players,*next_player;
	struct interactive *ip;

	char buff[TBUFF];
	int cs,start;

	eval_cost = 0;
        num_error=0;

        new_error=1;

	if (game_is_being_shut_down) shutdowngame(0); 

	if (slow_shut_down_to_do) 
	  {
	    char tmp = slow_shut_down_to_do;
	    slow_shut_down_to_do = 0;
	    slow_shut_down(tmp);
	  }

	purge_dead_sockets();

        if(!ob_status[0]) ob_status[0]=alloc_ob_status();

	check_for_connects();

	if(!all_players)
	    {
		extern int call_out_delay();
		int del=call_out_delay();

		if(del<0 || del>2) del=2;
		
		sleep(del);
	    }
	else 
              get_messages();

#ifndef AIX
	ftime(&wait_time);
#endif
    /*
     * Now proccess commands from the player.  The commands can go to
     * one of several places in the following order.
     *
     * If the command == "!!\n" then we repeate the last command.
     *
     * If it starts with a '!', then it is eather an escape from the
     * editor or from an input(), so we send it as a command to the
     * parser.
     *
     * if the player has a pending input() then the information is
     * sent to it.
     *
     * Otherwise, send the command to the parser.
     *
     * The player_parser() will find that current_object is 0, and
     * then set current_object to point to the object that defines
     * the command. This will enable such functions to be static.
     */

     if(!next_player) next_player=all_players;

     for(more_cmds=0,ip=next_player,current_time=time(0l);ip && current_time==time(0l);ip=next_player)
       {
	  extern char *first_cmd_in_buf PROT((struct interactive *));
	  extern void next_cmd_in_buf PROT((struct interactive *));
	  extern void write_file PROT((char *,char *,int));

	  char p[TBUFF],*buff=0;

	  next_player=ip->next;

  	  if(!(buff=first_cmd_in_buf(ip))) continue;

	/*
	 * we have a player cmd find out were it goes and send it there.
	 */
	
	command_giver = ip->ob;

        (void)strcpy(p,buff);

	next_cmd_in_buf(ip); /* move buffer pointers */

	if(!more_cmds && first_cmd_in_buf(ip)) 
	    more_cmds=1;
	
	/* manage snooping. */

	if (ip->snoop_by && !ip->noecho) {
	    command_giver = ip->snoop_by->ob;
	    add_message("%% %s\n", p);
	    command_giver = ip->ob;
	}

	/* if the object is logged, log it! */

	if(command_giver->logged) 
	  {
	    write_file(command_giver->logged,p,1);
	    write_file(command_giver->logged,"\n",1);
	  }

	if (ip->noecho) {
	    /* Must not enable echo before the user input is received. */
	    add_message("%c%c%c", IAC, WONT, TELOPT_ECHO);
	}

	ip->noecho = 0;
	ip->last_time = current_time;

        eval_cost=0;  /* this should be the cost per command */

    if(setjmp(error_recovery_context))
      {
	extern char case_match,break_flag,break_level;
#ifdef AIX
	extern signed char trace_depth;
#else
	extern char trace_depth;
#endif
	extern unsigned short current_line;
	extern struct lnode_2 *next_arg_list_to_use;
	extern LNODE *case_def;
	extern VALUE *switch_arg;

	/* if we get an error we need to do some house cleaning before
	 * we continue on.
	 */

	free_all_values();
	remove_destructed_objects();

	if(ob_status[0]) free_ob_status(ob_status[0]);
	ob_status[0]=alloc_ob_status();

	if(ob_status[1]) free_ob_status(ob_status[1]);

	command_giver=0;
	case_match=0;
	case_def=0;
	next_arg_list_to_use=0;
	break_flag=0;
	break_level=0;
	current_line=0;
	trace_depth=-1;
     }
    else
      {
	if (p[0] == '!')
	  {
	     if(p[1]=='!' && !p[2] && check_last_cmd())
		{
		     (void)strcpy(p,get_last_cmd());
		     parse_command(p,command_giver);
		}
	     else
	       {
	   	     set_last_cmd(p);
		     parse_command(p+1, command_giver);
	       }
	  }
	else if(!call_function_interactive(command_giver->interactive,p))
	  if (command_giver->ed_buffer)
	    {
	      int c;

	      for(c=0;c<cf.max_ed && command_giver->ed_buffer;c++)
	        {
		   set_last_cmd(p);
		   ed_cmd(p);
		   if(c<cf.max_ed-1 && command_giver->ed_buffer)
		     {
			if(buff=first_cmd_in_buf(ip))
		 	  {
			    next_cmd_in_buf(ip);
			    update_load_av();
			    (void)strcpy(p,buff);
			  }
			else break;
		     }
		   else break;
		}
	    }
	    else 
	      {
		     set_last_cmd(p);
		     parse_command(p, command_giver);
	      }

	    /*
	     * Print a prompt if player is still here.
	     */

	    if (command_giver && command_giver->interactive) print_prompt();

	    update_load_av();
          }
       }
     
    command_giver = 0;
    free_all_values();
    calls_heart_beats();
  }

/* we never come here. */
}

/*
 * Call all heart_beat() functions in all objects.  Also call the next reset,
 * and the call out.
 * We do heart beats by moving each object done to the end of the heart beat
 * list before we call its function, and always using the item at the head
 * of the list as our function to call.  We keep calling heart beats until
 * a timeout or we have done num_heart_objs calls.  It is done this way so
 * that objects can delete heart beating objects from the list from within
 * their heart beat without truncating the current round of heart beats.
 */
struct link_list *hb_list = 0; /* head */
static struct link_list *hb_start =0; /* second start */
static struct link_list *hb_begin =0; /* second end */

static int num_hb_objs = 0;  /* stats */
static double perc_hb_probes=100.0; /* decaying avge of how many complete */

void call_heart_beat() 
{

   if (hb_list)
     {  
	struct jmp_globals jmp;
	struct link_list *hb;
        int cs,num_done = 0,end_call,cur_num_hb=num_hb_objs;

	command_giver=0;
        set_jump_state(0,&jmp);
	ob_status[0]=0;
	ob_status[1]=0;

	if(current_heart_beat) hb=current_heart_beat;
		else hb=hb_list;

	if(hb!=hb_list) hb_start=hb_list;
		else hb_start=0;

	end_call=time(0l);

	hb_begin=hb;

        while (hb && end_call==time(0l))
           {
		jmp_buf erc;

	    	num_done++;
		current_heart_beat=hb->next;

       		ob_status[0]=alloc_ob_status();
    		ob_status[0]->ob = hb->ob;

    		if (hb->ob->flags & COMMANDS) command_giver = hb->ob;
      			else command_giver = 0;

    		hb->ob->wl->heart_beats++;
		eval_cost = 0;

		set_jump_state(erc,0);
		if(!setjmp(error_recovery_context))
		  {
			(void)apply(hb->fun_p,0,hb->ob,&const0,0);
       	        	free_ob_status(ob_status[0]);
	   	        if(ob_status[1]) free_ob_status(ob_status[1]);
			ob_status[0]=0;
			ob_status[1]=0;
		   }
		 else clear_jump_state(&jmp);
		 reset_jump_state(erc);

		eval_cost=0;
	  	if(current_heart_beat==hb_begin) break;
		hb=current_heart_beat;
		if(!hb && hb_start) hb=hb_start;
	   }

      if(!current_heart_beat) perc_hb_probes=100.0;
	else perc_hb_probes = 100 * (float) num_done / cur_num_hb;

      clear_jump_state(&jmp);
      free_all_values();
    }
}

/*
 * add or remove an object from the heart beat list; does the major check...
 * If an object removes something from the list from within a heart beat,
 * various pointers in call_heart_beat could be stuffed, so we must
 * check current_heart_beat and adjust pointers.
 */

int set_heart_beat(ob, to)
    struct object * ob;
    int to;
{
    struct link_list *new_hb;
    struct fun_p *fun_p;
    
    if ((to && ob->links[1]) || (!to && !ob->links[1]))
		return 0;

    if (to) 
	  {
		fun_p=find_fun_p("heart_beat",ob,0);

		if(!fun_p)
			error(ob->realm,"Tried to set heart_beat() in an object that does not have one defined!\n");

		new_hb=(struct link_list *)xalloc(sizeof (struct link_list));
		new_hb->fun_p=fun_p;
		new_hb->ob=ob;
		ob->links[1]=new_hb;
		new_hb->prev=0;
		if(hb_list) hb_list->prev=new_hb;
		new_hb->next = hb_list;
		hb_list = new_hb;
		num_hb_objs++;
      }
    else 
	 { /* remove all refs */

		if(ob->links[1]->prev)
		   ob->links[1]->prev->next=ob->links[1]->next;

		if(ob->links[1]->next)
			ob->links[1]->next->prev=ob->links[1]->prev;

		if(ob->links[1]==current_heart_beat)
		    current_heart_beat=ob->links[1]->next;

		if(ob->links[1]==hb_start) hb_start=ob->links[1]->next;

		if(hb_start==hb_begin) hb_start=0;

		if (hb_list == ob->links[1]) hb_list = ob->links[1]->next;

		if(ob->links[1]==hb_begin) hb_begin=ob->links[1]->next;

		free_fun_p(ob->links[1]->fun_p);

		free((char *)ob->links[1]);
		ob->links[1]=0;
		num_hb_objs--;
	 }
    return 1 ;
}

void heart_beat_status() 
{
    char buff[200];
    add_message("Heart beats                                 : %7d (%9d) bytes\n",
		num_hb_objs, num_hb_objs * sizeof (struct link_list));
    (void)sprintf(buff,"%7.2f%%",perc_hb_probes);
    add_message("Percentage of HB calls completed last time  : %s\n", buff);
}

void reset_status()
{
   add_message("Objects set to reset                        : %7d (%9d) bytes\n",
		num_resets_in_list,num_resets_in_list * sizeof (struct link_list));
   add_message("Number of resets                            : %7d\n",num_resets_called);
}


void sys_stat()
{
   VALUE *tmp=query_load_av();
   long hr,sec,min;
   long cpu_hr,cpu_frac,cpu_sec,cpu_min,cpu;
   double cpu_per;
   struct tms cpu_time;
   char buff[200];

   extern char *ctime();
#ifndef hp300
   extern long times();
#endif


   sec=time(0l)-start_time;
   (void)times(&cpu_time);

   cpu=(cpu_frac=cpu_time.tms_utime+cpu_time.tms_stime)/60;

   cpu_hr=cpu_frac/216000;
   cpu_frac-=cpu_hr*216000;
 
   cpu_min=cpu_frac/3600;
   cpu_frac-=cpu_min*3600;

   cpu_sec=cpu_frac/60;
   cpu_frac-=cpu_sec*60;

   cpu_per=((double)cpu/(double)sec)*100.0;

   hr=sec/3600;
   sec-=hr*3600;

   min=sec/60;
   sec-=min*60;

   add_message("Up since                                    : %s",ctime(&start_time));
   add_message("Time up                                     : %02d:%02d:%02d\n",hr,min,sec);
   add_message("Time on CPU                                 : %02d:%02d:%02d:%02d\n",cpu_hr,cpu_min,cpu_sec,cpu_frac);
   (void)sprintf(buff,"%.2f%%",cpu_per);
   add_message("Average CPU usage                           : %s\n",buff);
   add_message("Load average                                : %s\n",tmp->u.string);
}

/*
 * There is a file with a list of objects to be initialized at
 * start up.
 */


void load_first_objects() {
	struct jmp_globals jmp;
    FILE *f,*f2;
    char buff[100],r_buff[70],dir[15],name[40],owner[15],next_init_file[100];
    char *p;
    struct tms tms1, tms2;
 
    extern void load_mcp();

    next_init_file[0]=owner[0]=name[0]=dir[0]=0;
    f = fopen(cf.init_file, "r");
	if(f==NULL)
	 { if(errno == EMFILE) fatal("Max files open.\n");
	   return;
	 }

    load_realm=sys_realm;

    mcp.ob=0;
    new_error=0;
    load_mcp();

    f2= fopen("info/Realms","r");
	if(f2==NULL)
         {
           (void)fclose(f);
	   if(errno == EMFILE) fatal("Max files open.\n");
           return;}
	set_jump_state(0,&jmp);
	while(1)
	  {
	        new_error=0;	
		debug_message(load_realm,"\nPreloading objects from: %s\n\n",next_init_file[0]!=0?next_init_file:cf.init_file);
    	while(1)
	  	 {
			jmp_buf erc;

#ifndef hp300
			extern long times();
#endif

			if (fgets(buff, sizeof buff, f) == NULL) break;
			if (buff[0] == '#') continue;
			p = strchr(buff, '\n');
			if (p != 0) *p = 0;
			if (buff[0] == '\0') continue;
			new_error=0;
			debug_message(load_realm,"Preloading: %s", buff);
			eval_cost = 0;
			times(&tms1);
			set_jump_state(erc,0);
			if(!setjmp(error_recovery_context))
			  {
				new_error=1;
				(void)load_object(load_realm,buff);
				times(&tms2);
				new_error=0;
				debug_message(load_realm," %.2f\n", (tms2.tms_utime - tms1.tms_utime + tms2.tms_stime - tms1.tms_stime) / 60.0);
			  }
			 else clear_jump_state(&jmp);
			reset_jump_state(erc);
			free_all_values();
      	 }
        (void)fclose(f);
		f=0;
		while(1)
	      {
			if(fgets(r_buff,sizeof r_buff,f2)==NULL)
			 {
				(void)fclose(f2);
				f2=0;
				break;
			 }
 			(void)sscanf(r_buff,"%[^:]:%[^:]:%[^\n]\n",name,owner,dir);
			if(!strcmp(owner,"sysop")) continue;

			 (void)sprintf(next_init_file,"%s/%s",dir,cf.init_file);

	       		 f=fopen(next_init_file,"r");

  	                 if(!f)
		          if(errno == EMFILE) fatal("Max files open.\n");
				else continue;

		         load_realm=extract_realm(next_init_file);
			 new_error=0;
			 debug_message(sys_realm,"\nPreloading files in realm: %s\n",load_realm->realm_name);
			 debug_message(sys_realm,"From: %s\n",next_init_file);

			 break;
		  }
		if(!f || !f2) break;
	  }
    if(f2) (void)fclose(f2);
    if(f) (void)fclose(f);
}

int last_obj_clean_up=0;
void calls_heart_beats() {
	call_out();
	if(hb_time+1<time(0l))
	 {
	   hb_time=time(0l);
	   call_heart_beat();
	 }
         move_current_reset(0);
	if(!last_obj_clean_up) last_obj_clean_up=time(0l)+300;
	if(last_obj_clean_up<time(0l))
	    {
		  remove_destructed_objects();
		  last_obj_clean_up=time(0l)+300;
		}
}

struct link_list *reset_list=0;

int reset_delay=0;

void show_reset_list()
{
   struct link_list *tmp;
   int c;

   for(c=1,tmp=reset_list;tmp;tmp=tmp->next,c++)
	add_message("%3d. %s\n",c,tmp->ob->name);
}

void move_current_reset(flag) 
  int flag;
{
    struct value arg,*ret;
    extern int num_player;
    int i;
    struct link_list *current_reset;
    struct object *ob;
    
if(!reset_delay) reset_delay=MAX_RESET+time(0l);
if(!flag && reset_delay>time(0l)) return;
    reset_delay=MAX_RESET+time(0l)-(time(0l)-reset_delay);
    num_resets_called++;
    wiz_decay();
    save_wiz_file();
    if(reset_list)
      {
	struct jmp_globals jmp;

    set_jump_state(0,&jmp);
    for(current_reset=reset_list;current_reset;current_reset=next_reset)
     {
	jmp_buf erc;

	next_reset=current_reset->next;

	set_jump_state(erc,0);

         /* We have to catch an error here, locally.
          * It is not good if the error is catched globally, as the current
          * reset wouldn't be moved forward.
          */

	if(setjmp(error_recovery_context))
		clear_jump_state(&jmp);
        else
          {
  	    	if(current_reset->ob->flags & RESET)
	      	  {
		        current_reset->ob->not_touched=0;
			eval_cost=0;
                        (void)apply(current_reset->ob->reset,0, current_reset->ob, &const1,0);
	      	  }
		else current_reset->ob->not_touched++;
          }
	  reset_jump_state(erc);
	  free_all_values();
    }

   for(i=1;i<=num_player;i++)
	{
		jmp_buf erc;

		set_jump_state(erc,0);

         /* We have to catch an error here, locally.
          * It is not good if the error is catched globally, as the current
          * reset wouldn't be moved forward.
          */
        if (setjmp(error_recovery_context))
			clear_jump_state(&jmp);
	    else
		  {
			 extern OBJECT *get_interactive_object PROT((int,int));
			 ob=get_interactive_object(i,0);
			 if(ob) ob=ob->ref_ob[0];
			   else continue;
			for(;ob;ob=ob->ref_ob[1])
			 if(ob && ob->not_touched && ob->links[0])
			   {
			 	  ob->not_touched=0;
				  eval_cost=0;
			 	  (void)apply(ob->reset,0,ob,&const1,0);
			   }
		  }
	  reset_jump_state(erc);
	  free_all_values();
	}
  }
}

void remove_reset(link)
	struct link_list *link;
{

	if(link->prev)
		link->prev->next=link->next;

	if(link->next)
		link->next->prev=link->prev;

	if(link==next_reset) 
		next_reset=link->next;

	if(link==reset_list)
		reset_list=link->next;

	link->ob->links[0]=0;

	link->ob=0;
	link->next=0;
	link->prev=0;

	num_resets_in_list--;

	free((char *)link);
}

void set_reset(ob)
	OBJECT *ob;
{

	struct link_list *tmp=(struct link_list *)xalloc(sizeof (struct link_list));

	tmp->ob=ob;
	ob->links[0]=tmp;
	tmp->prev=0;
	if(reset_list) reset_list->prev=tmp;
	tmp->next=reset_list;
	reset_list=tmp;
	num_resets_in_list++;
}

extern double consts[5];
static double load_av = 0.0;

void update_load_av() {
    static unsigned long last_time;
    int i;
    long n;
    double c;
    static acc = 0;

    acc++;
    if (current_time == last_time)
	return;
    n = current_time - last_time;
    if (n < sizeof consts / sizeof consts[0])
	c = consts[n];
    else
	c = exp(- n / 900.0);
    load_av = c * load_av + acc * (1 - c) / n;
    last_time = current_time;
    acc = 0;
}

static double compile_av = 0.0;

void update_compile_av(lines)
    int lines;
{
    static unsigned long last_time;
    int i;
    long n;
    double c;
    static acc = 0;

    acc += lines;
    if (current_time == last_time)
	return;
    n = current_time - last_time;
    if (n < sizeof consts / sizeof consts[0])
	c = consts[n];
    else
	c = exp(- n / 900.0);
    compile_av = c * compile_av + acc * (1 - c) / n;
    last_time = current_time;
    acc = 0;
}

struct value *query_load_av() {
    char buff[100];

    sprintf(buff, "%.2f cmds/s, %.2f comp lines/s", load_av, compile_av);
    return make_string(buff);
}
