#include <stdio.h>
#include <time.h>
#include <string.h>
#include "config.h"
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "error.h"
#include "s_hash.h"
#include "realms.h"

/*
 * This file implements delayed calls of functions.
 * Static functions can not be called this way.
 *
 */

extern void free_svalue(),assign_svalue(),add_message();

extern char *make_shared_string PROT((char *));
extern char *xalloc(), *string_copy();

struct call {
    long delta;
    long delay;
    char *function;
    struct fun_p *fun_p;
    struct object *ob;
    struct svalue v;
    struct call *next;
};

static struct call *call_list=0;
static struct call *current_call=0;
static struct call *prev_current=0;
static int num_call=0;

int call_out_delay()
{
  int next_call= -1;

  if(call_list) 
     {
	next_call= call_list->delta-time(0l);
	if(next_call<0) next_call=0;
     }
  
  return next_call;
}

static struct call *call_free_list=0;

static struct call *alloc_call()
{
    register struct call *cop;

    if (!call_free_list) 
      {
	int i;

	call_free_list =(struct call *)xalloc(cf.alloc_call*sizeof (struct call));

	for (i=0; i<cf.alloc_call - 1; i++)
	    call_free_list[i].next  = &call_free_list[i+1];

	call_free_list[cf.alloc_call-1].next = 0;
    }

    cop = call_free_list;
    call_free_list = call_free_list->next;
    num_call++;

    return cop;
}

/*
 * Free a call out structure.
 */

static void free_call(cop)
    register struct call *cop;
{
    free_svalue(&cop->v);
    free_fun_p(cop->fun_p);
    free_string(cop->function,"free_call");
    num_call--;
    cop->next=call_free_list;
    call_free_list=cop;
}

/*
 * Setup a new call out.
 */
void new_call_out(ob, fun, delay, arg)
    register struct object *ob;
    register char *fun;
    int delay;
    register struct value *arg;
{
    register struct call *cop;
    struct fun_p *p;
   
    if (delay < 0) delay = 0;

    p=find_fun_p(fun,ob,0);

    if(!p)
       error(ob->realm,"call_out: '%s' function does not exist.\n",fun); 

    cop=alloc_call();

    cop->function = make_shared_string(fun);

    cop->ob = ob;

    cop->fun_p=p;
    if (arg) assign_svalue(&cop->v, arg,0);
      else 
	{
    	  cop->v.type = T_INT;
    	  cop->v.u.n_int = 0;
	}

    cop->delta=delay+time(0l);
    cop->delay=delay;

    if(call_list)
	 {
		struct call *copp,*prev=0;

		for(copp=call_list;copp;prev=copp,copp=copp->next)
	  	 {
			if(cop->delta>copp->delta)
			  {
			    if(!copp->next)
				{
				   copp->next=cop;
				   cop->next=0;
				   return;
				}
			    continue;
			  }
			cop->next=copp;
			if(prev) prev->next=cop;
			   else 
			    {
			       if(call_list==current_call)
				    prev_current=cop;
			       cop->next=call_list;
			       call_list=cop;
			    }
			return;
	  	 }
		free_call(cop);
		error(sys_realm,"bad call_list\n");
	  }
	cop->next=0;
	call_list=cop;
}

/*
 * See if there are any call outs to be called.
 */
void call_out() {
    register struct call *cop;
    extern struct object_status *ob_status[];
	extern void free_all_values();
	struct jmp_globals jmp;
	int start_time;
        extern int eval_cost;
	struct call *list;

    if (!call_list || call_list->delta>time(0l)) return;
		
    eval_cost=0;
    start_time=time(0l);
    prev_current=0;
    list=call_list;

    while (list && list->delta <=time(0l)) /* && start_time==time(0l)) */ 
	 {
    VALUE v;
	/*
	 * Move the first call_out out of the chain.
	 */
	current_call = cop = list;
	if (check_object(cop->ob))
	  {
	    /* We have to catch an error here, locally.
	     * It is not good if the error is catched globally, as the current
	     * call out wouldn't be removed.
	     */
		register OBJECT *ob_save=cop->ob;
		extern struct value const0;
		jmp_buf erc;
		
		if (cop->ob->flags & COMMANDS) command_giver = cop->ob;
			else command_giver = 0;

   		ob_status[0]=alloc_ob_status();
		ob_status[0]->ob=cop->ob;
		ob_status[1]=0;

		if (cop->v.type == T_OBJECT && !check_object(cop->v.u.ob))
			free_svalue(&cop->v);

		v.type = cop->v.type;
		v.u.ob = cop->v.u.ob;
		v.u.n_int=cop->v.u.n_int;
		v.u.n_float=cop->v.u.n_float;
		v.u.string=cop->v.u.string;
		v.u.vec=cop->v.u.vec;

		set_jump_state(erc,&jmp);
		if(!setjmp(error_recovery_context))
		        (void)apply(cop->fun_p,0, cop->ob, &v,0);
		else clear_jump_state(&jmp);
		reset_jump_state(erc);

	        free_ob_status(ob_status[0]);
	        if(ob_status[1]) free_ob_status(ob_status[1]);
		ob_status[0]=0;
		ob_status[1]=0;
		eval_cost=0;

        }
   
    list=cop->next;
    if(!prev_current) call_list = call_list->next;
      else {prev_current->next=cop->next;}

    free_call(cop);
   }
  current_call=prev_current=0;
  free_all_values();
}

/*
 * Throw away a call out. First call to this function is discarded.
 * The time left until execution is returned.
 * -1 is returned if no call out pending.
 */
 /* New third argument for internal use.  if(flag)  remove all call_outs
    in list from the object 'ob'.  'fun' is ignored as it would be 0.
  */

int remove_call_out(ob, fun,flag)
    struct object *ob;
    char *fun;
    int flag;
{
    struct call *copp, *cop,*prev=0;
    int delay;
    

    for(copp=call_list;copp;copp=cop)
  	  {
		cop=copp->next;

	 	if(copp!=current_call && 
		  (copp->ob==ob && (flag || !strcmp(copp->function,fun))))
		  {
		     if(prev) prev->next=copp->next;
		       else call_list=copp->next;
		     delay=copp->delta-time(0l);
		     if(delay<0) delay=0;
		     free_call(copp);
		     if(!flag) return delay;
		     continue;
		  }
		prev=copp;
	  }
    return -1;
}

VALUE *query_call_out(ob,fun)
    OBJECT *ob;
    char *fun;
{
   VALUE *ret;
   struct call *cop;

    ret=make_number(-1,0.0,0);
	cop=call_list;
	while(cop)
	 {
        if(cop->ob!=ob || strcmp(cop->function,fun)) 
			{cop=cop->next;continue;}
	   ret->u.n_int=cop->delta=time(0l);
	   if(ret->u.n_int<0) ret->u.n_int=0;
	   break;
	 }
    return ret;
}

void show_call_out_list()
{
  int count;
  struct call *cop;

  if(!call_list) return;
  for(cop=call_list,count=1;cop;cop=cop->next,count++)
    add_message("%3d. ob: %s\tfun: %s\tdelay:%d\ttime: %d\n",
	   count,cop->ob->name,cop->function,cop->delay,cop->delta-time(0l));
}
  
int print_call_out_usage() 
{
    add_message("Call outs                                   : %7d (%9d) bytes\n", num_call,
		num_call * sizeof (struct call));
    return num_call * sizeof (struct call);
}
