/*
 *  ed - standard editor
 *  ~~
 *	Authors: Brian Beattie, Kees Bot, and others
 *
 * Copyright 1987 Brian Beattie Rights Reserved.
 * Permission to copy or distribute granted under the following conditions:
 * 1). No charge may be made other than reasonable charges for reproduction.
 * 2). This notice must remain intact.
 * 3). No further restrictions may be added.
 * 4). Except meaningless ones.
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *  TurboC mods and cleanup 8/17/88 RAMontante.
 *  Further information (posting headers, etc.) at end of file.
 *  RE stuff replaced with Spencerian version, sundry other bugfix+speedups
 *  Ian Phillipps. Version incremented to "5".
 * _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
 *
 *  Genreal bug fixes and cleanups.  Added query_ed function.
 *  1/22/92 Man-ing.
 *  Version incremented to "6".
 */


#define ED_VER 0.7	/* used only in the "set" function, for i.d. */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "config.h"
#include "lint.h"
#include "regexp.h"

/* Regexp is Henry Spencer's package. WARNING: regsub is modified to return
 * a pointer to the \0 after the destination string, and this program refers
 * to the "private" reganch field in the struct regexp.
 */

#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "error.h"
#include "realms.h"

/*
 *	#defines for non-printing ASCII characters
 */
#define NUL	0x00	/* ^@ */
#define EOS	0x00	/* end of string */
#define SOH	0x01	/* ^A */
#define STX	0x02	/* ^B */
#define ETX	0x03	/* ^C */
#define EOT	0x04	/* ^D */
#define ENQ	0x05	/* ^E */
#define ACK	0x06	/* ^F */
#define BEL	0x07	/* ^G */
#define BS	0x08	/* ^H */
#define HT	0x09	/* ^I */
#define LF	0x0a	/* ^J */
#define NL	'\n'
#define VT	0x0b	/* ^K */
#define FF	0x0c	/* ^L */
#define CR	0x0d	/* ^M */
#define SO	0x0e	/* ^N */
#define SI	0x0f	/* ^O */
#define DLE	0x10	/* ^P */
#define DC1	0x11	/* ^Q */
#define DC2	0x12	/* ^R */
#define DC3	0x13	/* ^S */
#define DC4	0x14	/* ^T */
#define NAK	0x15	/* ^U */
#define SYN	0x16	/* ^V */
#define ETB	0x17	/* ^W */
#define CAN	0x18	/* ^X */
#define EM	0x19	/* ^Y */
#define SUB	0x1a	/* ^Z */
#define ESC	0x1b	/* ^[ */
#define FS	0x1c	/* ^\ */
#define GS	0x1d	/* ^] */
#define RS	0x1e	/* ^^ */
#define US	0x1f	/* ^_ */
#define SP	0x20	/* space */
#define DEL	0x7f	/* DEL*/
#define ESCAPE  '\\'


#define TRUE	1
#define FALSE	0
#define ERR		-2
#define FATAL		(ERR-1)
#define CHANGED		(ERR-2)
#define SET_FAIL	(ERR-3)
#define SUB_FAIL	(ERR-4)
#define MEM_FAIL	(ERR-5)

#define LGLOB	2       /* line marked global */

#define MAXLINE	 256	/* max number of chars per line */
#define MAXPAT	 256	/* max number of chars per replacement pattern */
#define MAXFNAME 100	/* max file name size */


/**  Global variables  **/

struct	line {
	int		l_stat;		/* empty, mark */
	struct line	*l_prev;
	struct line	*l_next;
	char		l_buff[1];
};
typedef struct line	LINE;

extern struct realms *load_realm;

extern void add_message();
extern struct object *command_giver;
extern struct object_status *ob_status[];
extern void set_prompt();
extern char *query_prompt();
char *query_ed();
extern int legal_path PROT((char *));
extern char *check_file_name PROT((char *,int,int));

void doprnt PROT((int, int,int)),clear_buffer PROT((struct ed_buffer *));
int ins PROT((char *));
int deflt PROT((int, int));

#define P_DIAG		(command_giver->ed_buffer->diag)
#define P_NONASCII	(command_giver->ed_buffer->nonascii)
#define P_NULLCHAR	(command_giver->ed_buffer->nullchar)
#define P_TRUNCATED	(command_giver->ed_buffer->truncated)
#define P_FNAME		(command_giver->ed_buffer->fname)
#define P_FCHANGED	(command_giver->ed_buffer->fchanged)
#define P_NOFNAME	(command_giver->ed_buffer->nofname)
#define P_READONLY 	(command_giver->ed_buffer->readonly)
#define P_MARK		(command_giver->ed_buffer->mark)
#define P_OLDPAT	(command_giver->ed_buffer->oldpat)
#define P_LINE0		(command_giver->ed_buffer->Line0)
#define P_CURLN		(command_giver->ed_buffer->CurLn)
#define P_CURPTR	(command_giver->ed_buffer->CurPtr)
#define P_LASTLN	(command_giver->ed_buffer->LastLn)
#define P_LINE1		(command_giver->ed_buffer->Line1)
#define P_LINE2		(command_giver->ed_buffer->Line2)
#define P_NLINES	(command_giver->ed_buffer->nlines)
#define P_NFLG		(command_giver->ed_buffer->nflg)
#define P_PFLG		(command_giver->ed_buffer->pflg)
#define P_APPENDING	(command_giver->ed_buffer->appending)
#define P_BLOCK_CMD	(command_giver->ed_buffer->block_cmd)

char	inlin[MAXLINE];
char	*inptr;		/* tty input buffer */
struct ed_buffer {
	int	nonascii;	/* count of non-ascii chars read */
	int	nullchar;	/* count of null chars read */
	int	truncated;	/* count of lines truncated */
	char fname[MAXFNAME];
	char saved_prompt[80];
	int	mark['z'-'a'+1];
	regexp	*oldpat;
	
	LINE	Line0;
	int	CurLn;
	LINE	*CurPtr;	/* CurLn and CurPtr must be kept in step */
	int	LastLn;
	int	Line1, Line2, nlines;
	char nflg;		/* print line number flag */
	char pflg;		/* witch mode to use when pflg flag is set. */
	char readonly;
	char nofname;
	char fchanged;	/* file-changed? flag */
	char appending;
        char block_cmd;
};

/*-------------------------------------------------------------------------*/

extern	char	*strcpy(), *strncpy();
extern	char	*xalloc();
extern	LINE	*getptr();
extern	char	*gettxt();
extern	char	*gettxtl();
extern	char	*catsub();
extern	void	prntln(), error();
regexp	*optpat();


/*________  Macros  ________________________________________________________*/

#ifndef max
#  define max(a,b)	((a) > (b) ? (a) : (b))
#endif

#ifndef min
#  define min(a,b)	((a) < (b) ? (a) : (b))
#endif

#ifndef toupper
#  define toupper(c)	((c >= 'a' && c <= 'z') ? c-32 : c )
#endif

#define nextln(l)	((l)+1 > P_LASTLN ? 0 : (l)+1)
#define prevln(l)	((l)-1 < 0 ? P_LASTLN : (l)-1)

#define gettxtl(lin)	((lin)->l_buff)
#define gettxt(num)	(gettxtl( getptr(num) ))

#define getnextptr(p)	((p)->l_next)
#define getprevptr(p)	((p)->l_prev)

#define setCurLn( lin )	( P_CURPTR = getptr( P_CURLN = (lin) ) )
#define nextCurLn()	( P_CURLN = nextln(P_CURLN), P_CURPTR = getnextptr( P_CURPTR ) )
#define prevCurLn()	( P_CURLN = prevln(P_CURLN), P_CURPTR = getprevptr( P_CURPTR ) )

#define clrbuf()	del(1, P_LASTLN)

#define	Skip_White_Space	{while (*inptr==SP || *inptr==HT) inptr++;}

#define relink(a, x, y, b) { (x)->l_prev = (a); (y)->l_next = (b); }


/*________  functions  ______________________________________________________*/


/*	append.c	*/


int append(line, glob)
int	line, glob;
{
	if(glob)
		return(ERR);
	setCurLn( line );
	P_APPENDING = 1;

	if(P_NFLG) add_message("%4d: ",P_CURLN+1);

	set_prompt("*\b",command_giver);
	return 0;
}

int more_append(str)
	char *str;
{
	if(str[0] == '.' && str[1] == '\0') {
		P_APPENDING = 0;
		add_message("\b\b\b\b\b\b\b");
	        doprnt(P_CURLN,P_CURLN,P_PFLG);
		set_prompt(":",command_giver);
		return(0);
	}

	if(P_NFLG) add_message("%4d: ",P_CURLN+2);

	if( ins(str) < 0) return( MEM_FAIL );
	return(0);
}

/*	ckglob.c	*/

int ckglob()
{
	regexp	*glbpat;
	unsigned char	c, delim;
	char	*lin;
	int	num;
	LINE	*ptr;

	c = *inptr;

	if(c != 'g' && c != 'v')
		return(0);
	if (deflt(1, P_LASTLN) < 0)
		return(ERR);

	delim = *++inptr;
	if(delim <= ' ')
		return(ERR);

	glbpat = optpat();
	if(*inptr == delim)
		inptr++;
	ptr = getptr(1);
	for (num=1; num<=P_LASTLN; num++) {
		ptr->l_stat &= ~LGLOB;
		if (P_LINE1 <= num && num <= P_LINE2) {
			lin = gettxtl(ptr);
			if(regexec(glbpat, lin )) {
				if (c=='g') ptr->l_stat |= LGLOB;
			} else  {
				if(reg_error) return(0);
				if (c=='v') ptr->l_stat |= LGLOB;
			}
		ptr = getnextptr(ptr);
		}
	}
	return(1);
}


/*  deflt.c
 *	Set P_LINE1 & P_LINE2 (the command-range delimiters) if the file is
 *	empty; Test whether they have valid values.
 */

int deflt(def1, def2)
int	def1, def2;
{
	if(!P_NLINES) {
		P_LINE1 = def1;
		P_LINE2 = def2;
	}
	return ( (P_LINE1>P_LINE2 || P_LINE1<=0) ? ERR : 0 );
}


/*	del.c	*/

/* One of the calls to this function tests its return value for an error
 * condition.  But del doesn't return any error value, and it isn't obvious
 * to me what errors might be detectable/reportable.  To silence a warning
 * message, I've added a constant return statement. -- RAM
 * ... It could check to<=P_LASTLN ... igp
 */

void del(from, to)
int	from, to;
{
	LINE	*first, *last, *next, *tmp;

	if(from < 1)
		from = 1;
	first = getprevptr( getptr( from ) );
	last = getnextptr( getptr( to ) );
	next = first->l_next;
	while(next != last && next != &P_LINE0) {
		tmp = next->l_next;
		free((char *)next);
		next = tmp;
	}
	relink(first, last, first, last);
	P_LASTLN -= (to - from)+1;
	setCurLn( prevln(from) );
}


/*	doprnt.c	*/

void doprnt(from, to,list)
int	from, to;
int list;
{
	from = (from < 1) ? 1 : from;
	to = (to > P_LASTLN) ? P_LASTLN : to;

	if(to != 0) {
		setCurLn( from );
		while( P_CURLN <= to ) {
			prntln( gettxtl( P_CURPTR ), (P_NFLG ? P_CURLN : 0),list);
			if( P_CURLN == to ) break;
			nextCurLn();
		}
	}
}


void prntln(str,lin,list)
char	*str;
int	list;
int	lin;
{
	if(lin) add_message("%4d: ",lin);
        if(list)
         {
	  while(*str && *str != NL) {
		if(*str < ' ' || *str >= 0x7f) {
			switch(*str) {
			case SOH:
				add_message("^A");
				break;
			case STX:
				add_message("^B");
				break;
			case ETX:
				add_message("^C");
				break;
			case EOT:
				add_message("^D");
				break;
			case ENQ:
				add_message("^E");
				break;
			case ACK:
				add_message("^F");
				break;
			case BEL:
				add_message("^G");
				break;
			case BS:
				add_message("^H");
				break;
			case HT:
				add_message("^I");
				break;
			case LF:
				add_message("^J");
				break;
			case VT:
				add_message("^K");
				break;
			case FF:
				add_message("^L");
				break;
			case CR:
				add_message("^M");
				break;
			case SO:
				add_message("^N");
				break;
			case SI:
				add_message("^O");
				break;
			case DLE:
				add_message("^P");
				break;
			case DC1:
				add_message("^Q");
				break;
			case DC2:
				add_message("^R");
				break;
			case DC3:
				add_message("^S");
				break;
			case DC4:
				add_message("^T");
				break;
			case NAK:
				add_message("^U");
				break;
			case SYN:
				add_message("^V");
				break;
			case ETB:
				add_message("^W");
				break;
			case CAN:
				add_message("^X");
				break;
			case EM:
				add_message("^Y");
				break;
			case SUB:
				add_message("^Z");
				break;
			case ESC:
				add_message("^[");
				break;
			case FS:
				add_message("^\\");
				break;
			case GS:
				add_message("^]");
				break;
			case RS:
				add_message("^^");
				break;
			case US:
				add_message("^_");
				break;
			case DEL:
			    add_message("^?");
				break;
			default:
				add_message("\\%o",*str);
				break;
			}
		} 
		else add_message("%c", *str);
		str++;
	}
     add_message("$\n");
   }
 else add_message("%s\n",str);
}

/*	egets.c	*/

int egets(str,size,stream)
char	*str;
int	size;
FILE	*stream;
{
	int	c, count;
	char	*cp;

	for(count = 0, cp = str; size > count;) {
		c = getc(stream);
		if(c == EOF) {
			*cp = EOS;
			if(count)
				add_message("[Incomplete last line]\n");
			return(count);
		}
		else if(c == NL) {
			*cp = EOS;
			return(++count);
		}
		else if (!c) P_NULLCHAR++;                    /* count nulls */
			else
		      {  if(c & 128)
				  {
				    P_NONASCII++;   /* count eight-bit ascii */
				 	c&=(char)~128;
				  }
				 *cp++ =(char)c;	/* not null, keep it */
				 count++;
		      }
	}
	str[count-1] = EOS;
	if(c != NL) {
		P_TRUNCATED++;
		while((c = getc(stream)) != EOF)
			if(c == NL) break;
	}
	return(count);
}  /* egets */


int doread(lin, fname)
int	lin;
char	*fname;
{
	FILE	*fp;
	int	err;
	unsigned long	bytes;
	int	lines;
	static char	str[MAXLINE];

	err = 0;
	P_NONASCII = P_NULLCHAR = P_TRUNCATED = 0;

	add_message("\"/%s\" ",fname);
	if( (fp = fopen(fname, "r")) == NULL ) {
	        if(!fp && errno == EMFILE) fatal("Max files open.\n");
		add_message(" isn't readable.\n");
		return( ERR );
	}
	setCurLn( lin );
	for(lines = 0, bytes = 0;(err = egets(str,MAXLINE,fp)) > 0;) {
		bytes += err;
		if(ins(str) < 0) {
			err = MEM_FAIL;
			break;
		}
		lines++;
	}
	fclose(fp);
	if(err < 0) return(err);
	add_message("%d lines %u bytes",lines,bytes);
	if(P_READONLY) add_message(" [Read Only]");
	if(P_NONASCII) add_message(" [%d eight-bit ascii]",P_NONASCII);
	if(P_NULLCHAR) add_message(" [%d nul]",P_NULLCHAR);
	if(P_TRUNCATED) add_message(" [%d lines truncated]",P_TRUNCATED);
	add_message("\n\n");
        setCurLn(1);
	if(deflt(P_CURLN,P_CURLN) >= 0)
		doprnt(P_LINE1,P_LINE2,0);
	return( err );
}  /* doread */


int dowrite(from, to, fname, apflg)
int	from, to;
char	*fname;
int	apflg;
{
	FILE	*fp;
	int	lin, err;
	int	lines;
	unsigned long	bytes;
	char	*str;
	LINE	*lptr;

	err = 0;
	lines = 0;
        bytes = 0;

	add_message("\"/%s\" ",fname);
	if((fp = fopen(fname,(apflg?"a":"w"))) == NULL) {
	        if(!fp && errno == EMFILE) fatal("Max files open.\n");
		add_message(" can't be opened for writing!\n");
		return( ERR );
	}

	lptr = getptr(from);
	for(lin = from; lin <= to; lin++) {
		str = lptr->l_buff;
		lines++;
		bytes += strlen(str) + 1;	/* str + '\n' */
		if(fputs(str, fp) == EOF) {
			add_message("file write error\n");
			err++;
			break;
		}
		fputc('\n', fp);
		lptr = lptr->l_next;
	}
	(void)fclose(fp);
	add_message("%d lines %lu bytes\n",lines,bytes);
	return( err );
}  /* dowrite */


/*	find.c	*/

int find(pat, dir)
regexp	*pat;
int	dir;
{
	int	i, num;
	LINE	*lin;

	if(dir) num = nextln(P_CURLN), lin=getnextptr(P_CURPTR);
	 else num=prevln(P_CURLN), lin = getprevptr(P_CURPTR);
	for(i=0; i<P_LASTLN; i++ ) {
		if(regexec( pat, gettxtl( lin ) ))
			return(num);
		if(reg_error) return(0);
		if( dir )
			num = nextln(num), lin = getnextptr(lin);
		else
			num = prevln(num), lin = getprevptr(lin);
	}
	return ( ERR );
}


/*	getfn.c	*/

char *getfn(writeflg)
    int writeflg;
{
	static char	file[MAXFNAME];
	char	*cp;
	char *file2;
	VALUE v,*v2,v3;

	if(*inptr == NL) 
	   {
		 if(P_FNAME[0]==0) return(NULL);
		 (void)sprintf(file,"/%s",P_FNAME);
		 P_NOFNAME=TRUE;
	   }
	  else
	  {
		P_NOFNAME=FALSE;
		Skip_White_Space;

		cp = file;
		while(*inptr && *inptr != NL && *inptr != SP && *inptr != HT)
			*cp++ = *inptr++;
		*cp = '\0';
		v.type=T_STRING;
		v.u.string=file;
		v3.type=T_STRING;
		v3.u.string=command_giver->realm->realm_dir;

		v2=apply(mcp.fun_p[2],0,mcp.ob,&v,&v3);
		if(strchk(v2,1) || !legal_path(v2->u.string)) 
		  {
			add_message("Illegal file name.\n");
			return( NULL );
	 	  }
		(void)strncpy(file,v2->u.string,strlen(v2->u.string)<MAXFNAME-3?strlen(v2->u.string):MAXFNAME-3);
		file[strlen(v2->u.string)<MAXFNAME-3?strlen(v2->u.string):MAXFNAME-2]=0;
		
	  }
	if(!(file2=check_file_name(file, 0,0)))
	  {
	    add_message("Bad file name.\n");
	    return( NULL );
	  }
	if(!check_file_name(file,1,0))
	  {
		if(writeflg)
	 	  {
	            add_message("[Read Only]\n");
		    return( NULL );
		  }
		P_READONLY=TRUE;
	  }
	else P_READONLY=FALSE;
	(void)strncpy(file, file2, strlen(file2)<MAXFNAME-3?strlen(file2):MAXFNAME-3);
	file[strlen(file2)<MAXFNAME-3?strlen(file2):MAXFNAME-2] = 0;
	if(!strlen(file)) {
		add_message("no file name\n");
		return(NULL);
	}
	return( file );
}  /* getfn */


int getnum(first)
int first;
{
	regexp	*srchpat;
	int	num;
	unsigned char	c;

	Skip_White_Space;

	if(*inptr >= '0' && *inptr <= '9') {	/* line number */
		for(num = 0; *inptr >= '0' && *inptr <= '9'; ++inptr) {
			num = (num * 10) + (*inptr - '0');
		}
		return num;
	}

	switch(c = *inptr) {
	case '.':
		inptr++;
		return (P_CURLN);

	case '$':
		inptr++;
		return (P_LASTLN);

	case '/':
	case '?':
		srchpat = optpat();
		if(*inptr == c)
			inptr++;
		return(find(srchpat,c == '/'?1:0));

	case '-':
	case '+':
		return(first ? P_CURLN : 1);

	case '\'':
		inptr++;
		if (*inptr < 'a' || *inptr > 'z')
			return(EOF);
		return P_MARK[ *inptr++ - 'a' ];

	default:
		return ( first ? EOF : 1 );	/* unknown address */
	}
}  /* getnum */


/*  getone.c
 *	Parse a number (or arithmetic expression) off the command line.
 */
#define FIRST 1
#define NOTFIRST 0

int getone()
{
	int	c, i, num;

	if((num = getnum(FIRST)) >= 0) {
		for (;;) {
			Skip_White_Space;
			if(*inptr != '+' && *inptr != '-')
				break;	/* exit infinite loop */

                        c = *inptr++;
			if((i = getnum(NOTFIRST)) < 0)
				return ( i );
			if(c == '+')
				num += i;
			else
				num -= i;
		}
	}
	return ( num>P_LASTLN ? ERR : num );
}  /* getone */


int getlst()
{
	int	num;

	P_LINE2 = 0;
	for(P_NLINES=0; (num = getone()) >= 0;)
	{
		P_LINE1 = P_LINE2;
		P_LINE2 = num;
		P_NLINES++;
		if(*inptr != ',' && *inptr != ';')
			break;
		if(*inptr == ';')
			setCurLn( num );
		inptr++;
	}
	P_NLINES = min(P_NLINES, 2);
	if(!P_NLINES)
		P_LINE2 = P_CURLN;
	if(P_NLINES <= 1)
		P_LINE1 = P_LINE2;

	return ( (num == ERR) ? num : P_NLINES );
}  /* getlst */


/*	getptr.c	*/

LINE *getptr(num)
int	num;
{
	LINE	*ptr;
	int	j;

	if (2*num>P_LASTLN && num<=P_LASTLN) {	/* high line numbers */
		ptr = P_LINE0.l_prev;
		for (j = P_LASTLN; j>num; j--)
			ptr = ptr->l_prev;
	} else {				/* low line numbers */
		ptr = &P_LINE0;
		for(j = 0; j < num; j++)
			ptr = ptr->l_next;
	}
	return(ptr);
}


/*	getrhs.c	*/

int getrhs(sub)
char	*sub;
{
	char delim = *inptr++;
	char *outmax = sub + MAXPAT;

	if( delim == NL) return( ERR );
        if(*inptr == NL) {*sub='\0';return( 0 );}	/* check for eol */

	while( *inptr != delim && *inptr != NL ) {
		if ( sub > outmax )
			return ERR;
		if ( *inptr == ESCAPE ) {
			switch ( *++inptr ) {
			case 'r':
				*sub++ = '\r';
				inptr++;
				break;
			case ESCAPE:
				*sub++ = ESCAPE;
				*sub++ = ESCAPE;
				inptr++;
			case 'n':
				*sub++ = '\n';
				inptr++;
				break;
			case 'b':
				*sub++ = '\b';
				inptr++;
				break;
			case 'f':
				*sub++='\f';
				inptr++;
				break;
			case 't':
				*sub++='\t';
				inptr++;
				break;
			case '0': {
				int i=3;
				*sub = 0;
				do {
					if (*++inptr<'0' || *inptr >'7')
						break;
					*sub = (*sub<<3) | (*inptr-'0');
				} while (--i!=0);
				sub++;
				} break;
			default:
				if ( *inptr != delim )
					*sub++ = ESCAPE;
				*sub++ = *inptr;
				if ( *inptr != NL )
					inptr++;
			}
		}
		else *sub++ = *inptr++;
	}
	*sub = '\0';

	inptr++;		/* skip over delimter */
	Skip_White_Space;
	if(*inptr == 'g') {
		inptr++;
		return( 1 );
	}
	return( 0 );
}

/*	ins.c	*/

int ins(str)
char	*str;
{
	char	*cp;
	LINE	*new, *nxt;
	int	len;

	do {
		for ( cp = str; *cp && *cp != NL; cp++ )
			;
		len = cp - str;
		/* cp now points to end of first or only line */

		if((new = (LINE *)xalloc(sizeof(LINE)+len)) == NULL)
			return( MEM_FAIL ); 	/* no memory */

		new->l_stat=0;
		strncpy(new->l_buff,str,len);	/* build new line */
		new->l_buff[len] = EOS;
		nxt = getnextptr(P_CURPTR);	/* get next line */
		relink(P_CURPTR, new, new, nxt);	/* add to linked list */
		relink(new, nxt, P_CURPTR, new);
		P_LASTLN++;
		P_CURLN++;
		P_CURPTR = new;
		str = cp + 1;
	}
		while( *cp != EOS );
	return 1;
}


/*	join.c	*/

int join(first, last)
int first, last;
{
	char buf[MAXLINE];
	char *cp=buf, *str;
	LINE *lin;
	int num;

	if (first<=0 || first>last || last>P_LASTLN)
		return(ERR);
	if (first==last) {
		setCurLn( first );
		return 0;
	}
	lin = getptr(first);
	for (num=first; num<=last; num++) {
		str=gettxtl(lin);
		while ( *str ) {
			if (cp >= buf + MAXLINE-1 ) {
				add_message("line too long\n");
				return(ERR);
			}
			*cp++ = *str++;
		}
		lin = getnextptr(lin);
	}
	*cp = EOS;
	del(first, last);
	if( ins(buf) < 0 )
		return MEM_FAIL;
	P_FCHANGED = TRUE;
	return 0;
}


/*  move.c
 *	Unlink the block of lines from P_LINE1 to P_LINE2, and relink them
 *	after line "num".
 */

int move(num)
int	num;
{
	int	range;
	LINE	*before, *first, *last, *after;

	if( P_LINE1 <= num && num <= P_LINE2 )
		return( ERR );
	range = P_LINE2 - P_LINE1 + 1;
	before = getptr(prevln(P_LINE1));
	first = getptr(P_LINE1);
	last = getptr(P_LINE2);
	after = getptr(nextln(P_LINE2));

	relink(before, after, before, after);
	P_LASTLN -= range;	/* per AST's posted patch 2/2/88 */
	if (num > P_LINE1)
		num -= range;

	before = getptr(num);
	after = getptr(nextln(num));
	relink(before, first, last, after);
	relink(last, after, before, first);
	P_LASTLN += range;	/* per AST's posted patch 2/2/88 */
	setCurLn( num + range );
	return( 1 );
}


int transfer(num)
int num;
{
	int mid, lin, ntrans;

	if (P_LINE1<=0 || P_LINE1>P_LINE2)
		return(ERR);

	mid= num<P_LINE2 ? num : P_LINE2;

	setCurLn( num );
	ntrans=0;

	for (lin=P_LINE1; lin<=mid; lin++) {
		if( ins(gettxt(lin)) < 0 )
			return MEM_FAIL;
		ntrans++;
	}
	lin+=ntrans;
	P_LINE2+=ntrans;

	for ( ; lin <= P_LINE2; lin += 2 ) {
		if( ins(gettxt(lin)) < 0 )
			return MEM_FAIL;
		P_LINE2++;
	}
	return(1);
}


/*	optpat.c	*/

regexp *optpat()
{
	char	delim, str[MAXPAT], *cp;

	delim = *inptr++;
	if (delim == NL)
	    return P_OLDPAT;
	cp = str;
	while(*inptr != delim && *inptr != NL && *inptr != EOS && cp < str + MAXPAT - 1) {
		if(*inptr == ESCAPE && inptr[1] != NL)
			*cp++ = *inptr++;
		*cp++ = *inptr++;
	}

	*cp = EOS;
	if(*str == EOS)
		return(P_OLDPAT);
	if(P_OLDPAT)
		free((char *)P_OLDPAT);
	return P_OLDPAT = regcomp(str);
}

int do_help()
{
	extern VALUE *print_file();

	if(strncmp(inptr,"elp",3)) return(ERR);
        inptr+=3;

	if(*inptr!=SP && *inptr !=HT && *inptr !=NL)
	   return(ERR);

	if(*inptr==NL) 
           {
            add_message("To View Help files Type:\n\n");
	    add_message("Help 1  -  Ed Commands.\n");
	    add_message("Help 2  -  Patern Expression Examples.\n");
	    add_message("Help 3  -  How to Search.\n");
	    add_message("Help 4  -  How to Copy from another File.\n\n");
           }
	  else if(*(++inptr)=='1') (void)print_file("/doc/ed/ed1",0,0,0);
            else if(*inptr=='2') (void)print_file("/doc/ed/ed2",0,0,0);
               else if(*inptr=='3') (void)print_file("/doc/ed/ed3",0,0,0);
                  else if(*inptr=='4') (void)print_file("/doc/ed/ed4",0,0,0);
			else return(ERR);

  doprnt(P_CURLN,P_CURLN,P_PFLG);
  return(0);
}


int set()
{
	char	word[16];
	int	i;

        if(*(++inptr)=='t') inptr++;

	if(*inptr != SP && *inptr != HT && *inptr != NL)
			return(ERR);

	if ( (*inptr == NL))
	{
		add_message("ed version %-3.1f\n\n",ED_VER);
		add_message("Line numbering     : %s\n", P_NFLG?"on":"off");
		add_message("Defualt Print mode : %s\n\n",P_PFLG?"list":"print");
	        doprnt(P_CURLN,P_CURLN,P_PFLG);
		return(0);
	}

	Skip_White_Space;
	for(i = 0; *inptr != SP && *inptr != HT && *inptr != NL;)
		word[i++] = *inptr++;
	word[i] = EOS;
	if(!strcmp(word,"number") || !strncmp(word,"line",4))
	{
	    if(P_NFLG) P_NFLG=FALSE;
		else P_NFLG=TRUE;
	}
       else if(!strncmp(word,"print",5))
	      {
		if(P_PFLG) P_PFLG=FALSE;
	      }
	    else if(!strcmp(word,"list"))
		    {
			if(!P_PFLG) P_PFLG=TRUE;
		    }
	          else return SET_FAIL;

	doprnt(P_CURLN,P_CURLN,P_PFLG);
	return(0);
}

#ifndef relink
void relink(a, x, y, b)
LINE	*a, *x, *y, *b;
{
	x->l_prev = a;
	y->l_next = b;
}
#endif


void set_ed_buf()
{
	relink(&P_LINE0, &P_LINE0, &P_LINE0, &P_LINE0);
	P_CURLN = P_LASTLN = 0;
	P_CURPTR = &P_LINE0;
}


/*	subst.c	*/

int subst(pat, sub, gflg, pflag)
regexp	*pat;
char	*sub;
int	gflg, pflag;
{
	int	nchngd = 0;
	char	*txtptr;
	char	*new, buf[MAXLINE];
	int	still_running = 1;
	LINE	*lastline = getptr( P_LINE2 );

	if(P_LINE1 <= 0)
		return( SUB_FAIL );
	nchngd = 0;		/* reset count of lines changed */

	for( setCurLn( prevln( P_LINE1 ) ); still_running; ) {
		nextCurLn();
		new = buf;
		if ( P_CURPTR == lastline )
			still_running = 0;
		if ( regexec( pat, txtptr = gettxtl( P_CURPTR ) ) ) {
			do
				{
				/* Copy leading text */
				int diff = pat->startp[0] - txtptr;
				strncpy( new, txtptr, diff );
				new += diff;
				/* Do substitution */
				new = regsub( pat, sub, new );
				txtptr = pat->endp[0];
				}
			while( gflg && !pat->reganch && regexec( pat, txtptr ));

			/* Copy trailing chars */
			while( *txtptr ) *new++ = *txtptr++;

			if(new >= buf+MAXLINE)
				return( SUB_FAIL );
			*new++ = EOS;
			del(P_CURLN,P_CURLN);
			if( ins(buf) < 0 )
				return MEM_FAIL;
			nchngd++;
			if(pflag)
				doprnt(P_CURLN, P_CURLN,0);
		}
		else if(reg_error) return(0);
        }
	return (( !nchngd && !gflg ) ? SUB_FAIL : nchngd);
}

/*  docmd.c
 *	Perform the command specified in the input buffer, as pointed to
 *	by inptr.  Actually, this finds the command letter first.
 */

int docmd(glob)
int	glob;
{
	static char	rhs[MAXPAT];
	regexp	*subpat;
	int	c, err, line3;
	int	apflg, pflag, gflag;
	int	nchng;
	char	*fptr;

	pflag = FALSE;
	Skip_White_Space;

	c = *inptr++;
	switch(c) {
	case NL:
		if(!P_NLINES && (P_LINE2 = nextln(P_CURLN)) == 0 )
			return(ERR);
		setCurLn( P_LINE2 );
		return (1);

	case '=':
		add_message("%d lines.\n",P_LASTLN);
		break;

	case 'n':
	  {
		char set_nflg=0;

		if(*inptr!=NL) return(ERR);
		if(deflt(P_CURLN,P_CURLN)<0) return(ERR);
		if(!P_NFLG) { P_NFLG=TRUE; set_nflg=1;}
		doprnt(P_CURLN,P_CURLN,P_PFLG);
	    if(set_nflg) P_NFLG=FALSE;
		break;
	  }
	case 'a':
		if(*inptr != NL || P_NLINES > 1)
			return(ERR);

		if(append(P_LINE1, glob) < 0)
			return(ERR);
		P_FCHANGED = TRUE;
		break;

	case 'c':
		if(*inptr != NL)
			return(ERR);

		if(deflt(P_CURLN, P_CURLN) < 0)
			return(ERR);

		del(P_LINE1, P_LINE2);
		if(append(P_CURLN, glob) < 0)
			return(ERR);
		P_FCHANGED = TRUE;
		break;

	case 'd':
		if(*inptr != NL)
			return(ERR);

		if(deflt(P_CURLN, P_CURLN) < 0)
			return(ERR);

		del(P_LINE1, P_LINE2);
		if(nextln(P_CURLN) != 0)
			nextCurLn();
		P_FCHANGED = TRUE;
		break;

	case 'e':
		if(P_NLINES > 0)
			return(ERR);
		if(P_FCHANGED)
			return CHANGED;
		/*FALL THROUGH*/
	case 'E':
		if(P_NLINES > 0)
			return(ERR);

		if(*inptr != ' ' && *inptr != HT && *inptr != NL)
			return(ERR);

		if((fptr = getfn(0)) == NULL)
			return(ERR);

		clrbuf();
		(void)doread(0, fptr);

		strcpy(P_FNAME, fptr);
		P_FCHANGED = FALSE;
		break;

	case 'f':
		if(P_NLINES > 0)
			return(ERR);

		if(*inptr != ' ' && *inptr != HT && *inptr != NL)
			return(ERR);

		if((fptr = getfn(1)) == NULL && !P_READONLY)
			return(0);

		if (P_NOFNAME) add_message("/%s %d lines.\n", P_FNAME,P_LASTLN);
			else if(!P_READONLY) (void)strcpy(P_FNAME,fptr);
		break;

	case 'i':
		if(*inptr != NL || P_NLINES > 1)
			return(ERR);

		if(append(prevln(P_LINE1), glob) < 0)
			return(ERR);
		P_FCHANGED = TRUE;
		break;

	case 'j':
		if (*inptr != NL || deflt(P_CURLN, P_CURLN+1)<0)
			return(ERR);

		if (join(P_LINE1, P_LINE2) < 0)
			return(ERR);
		break;

	case 'k':
		Skip_White_Space;

		if (*inptr < 'a' || *inptr > 'z')
			return ERR;
		c= *inptr++;

		if(*inptr != ' ' && *inptr != HT && *inptr != NL)
			return(ERR);

		P_MARK[c-'a'] = P_LINE1;
		break;

	case 'l':
		if(*inptr != NL)
			return(ERR);
		if(deflt(P_CURLN,P_CURLN) < 0)
			return(ERR);
		doprnt(P_LINE1,P_LINE2,1);
		break;

	case 'm':
		if((line3 = getone()) < 0)
			return(ERR);
		if(deflt(P_CURLN,P_CURLN) < 0)
			return(ERR);
		if(move(line3) < 0)
			return(ERR);
		P_FCHANGED = TRUE;
		break;

	case 'P':
	case 'p':
		if(*inptr != NL)
			return(ERR);
		if(deflt(P_CURLN,P_CURLN) < 0)
			return(ERR);
		doprnt(P_LINE1,P_LINE2,0);
		break;

	case 'q':
		if(P_FCHANGED)
			return CHANGED;
		/*FALL THROUGH*/
	case 'Q':
		clrbuf();
		if(*inptr == NL && !P_NLINES && !glob)
			return(EOF);
		else
			return(ERR);

	case 'r':
		if(P_NLINES > 1)
			return(ERR);

		if(!P_NLINES)			/* The original code tested */
			P_LINE2 = P_LASTLN;		/*	if(P_NLINES = 0)	    */
						/* which looks wrong.  RAM  */

		if(*inptr != ' ' && *inptr != HT && *inptr != NL)
			return(ERR);

		if((fptr = getfn(0)) == NULL)
			return(ERR);

		if((err = doread(P_LINE2, fptr)) < 0)
			return(err);
		P_FCHANGED = TRUE;
		break;

	case 'h':
		if(strncmp(inptr,"elp",3)) return(ERR);

		return(do_help());
		break;
	case 's':
		if(*inptr == 'e')
			return(set());
		Skip_White_Space;
		if((subpat = optpat()) == NULL)
			return(ERR);
		if((gflag = getrhs(rhs)) < 0)
			return(ERR);
		if(*inptr == 'p')
			pflag++;
		if(deflt(P_CURLN, P_CURLN) < 0)
			return(ERR);
		if((nchng = subst(subpat, rhs, gflag, pflag)) < 0)
			return(ERR);
		if(nchng)
			P_FCHANGED = TRUE;
		if (!P_BLOCK_CMD) doprnt(P_LINE1,P_LINE2,P_PFLG);
		break;

	case 't':
		if((line3 = getone()) < 0)
			return(ERR);
		if(deflt(P_CURLN,P_CURLN) < 0)
			return(ERR);
		if(transfer(line3) < 0)
			return(ERR);
		P_FCHANGED = TRUE;
		break;

	case 'W':
	case 'w':
		apflg = (c=='W');

		if(*inptr != ' ' && *inptr != HT && *inptr != NL)
			return(ERR);

		if((fptr = getfn(1)) == NULL)
			return(ERR);

		if(deflt(1, P_LASTLN) < 0)
			return(ERR);
		if(dowrite(P_LINE1, P_LINE2, fptr, apflg) < 0)
			return(ERR);
		P_FCHANGED = FALSE;
		break;

	case 'x':
		if(*inptr == NL && !P_NLINES && !glob) {
			if((fptr = getfn(1)) == NULL)
				return(ERR);
			if(dowrite(1, P_LASTLN, fptr, 0) >= 0)
				return(EOF);
		}
		return(ERR);

	case 'z':
		if(deflt(P_CURLN,P_CURLN) < 0)
			return(ERR);

		switch(*inptr) {
		case '-':
			doprnt(P_LINE1-21,P_LINE1,P_PFLG);
			break;

		case '.':
			doprnt(P_LINE1-11,P_LINE1+10,P_PFLG);
			break;

		case '+':
		case '\n':
			doprnt(P_LINE1,P_LINE1+21,P_PFLG);
			break;
		}
		break;

	default:
		return(ERR);
	}
	return (0);
}  /* docmd */


/*	doglob.c	*/
int doglob()
{
	int	lin, stat;
	char	*cmd;
	LINE	*ptr;

	cmd = inptr;

	P_BLOCK_CMD=1;

	for (;;) {
		ptr = getptr(1);
		for (lin=1; lin<=P_LASTLN; lin++) {
			if (ptr->l_stat & LGLOB)
				break;
			ptr = getnextptr(ptr);
		}
		if (lin > P_LASTLN)
			break;

		ptr->l_stat &= ~LGLOB;
		P_CURLN = lin; P_CURPTR = ptr;
		inptr = cmd;
		if((stat = getlst()) < 0)
			{P_BLOCK_CMD=0;return(stat);}
		if((stat = docmd(1)) < 0)
			{P_BLOCK_CMD=0;return(stat);}
	}
	P_BLOCK_CMD=0;
	return(P_CURLN);
}  /* doglob */


void ed_start(file_arg)
	char *file_arg;
{
	char *new_file;

	if (command_giver->ed_buffer)
		error(command_giver->realm,"Tried to start an ed session, when already active.\n");
	if (command_giver != ob_status[0]->ob)
	    	error(command_giver->realm,"Illegal start of ed.\n");
	command_giver->ed_buffer =
		(struct ed_buffer *)xalloc(sizeof (struct ed_buffer));
	clear_buffer(command_giver->ed_buffer);
	strcpy(command_giver->ed_buffer->saved_prompt,query_prompt(command_giver));
	command_giver->ed_buffer->CurPtr = &command_giver->ed_buffer->Line0;
	P_PFLG=0;
	P_NFLG=0;
	P_FNAME[0]=0;
	set_ed_buf();

	set_prompt(":",command_giver);
	add_message("Type 'Help' for a listing of ed() help commands.\n\n");
	if(file_arg)
	  {
		inptr=file_arg;
		if((new_file=getfn(0))!=NULL)
		  {
			(void)strncpy(P_FNAME,new_file,strlen(new_file)<MAXFNAME-3?strlen(new_file):MAXFNAME-3);
			P_FNAME[strlen(new_file)<MAXFNAME-3?strlen(new_file):MAXFNAME-2]=0;
			(void)doread(0,P_FNAME);
		   }
		  else
		   {
			 P_FNAME[0]=0;
			 setCurLn(1);
		   }
	  }
	else 
	  {
		P_FNAME[0]=0;
		setCurLn(1);
	  }

}

void ed_cmd(str)
	char *str;
{
	int stat;

	if (P_APPENDING) {
		more_append(str);
		return;
	}
	if (strlen(str) < MAXLINE)
	    strcat(str, "\n");
	
	strncpy(inlin, str, MAXLINE-1);
	inlin[MAXLINE-1] = 0;
	inptr = inlin;
	if(getlst() >= 0)
		if((stat = ckglob()) != 0) {
			if(stat >= 0 && (stat = doglob()) >= 0) {
				setCurLn( stat );
				return;
			}
		} else {
			if((stat = docmd(0)) >= 0) {
				if(stat == 1)
					doprnt(P_CURLN, P_CURLN,P_PFLG);
				return;
			}
		}
	switch (stat) {
	case EOF:
		set_prompt(command_giver->ed_buffer->saved_prompt,command_giver);
		if(P_NLINES) clrbuf();

		free((char *)command_giver->ed_buffer);
		command_giver->ed_buffer = 0;
		add_message("Exit from ed.\n");
		return;
	case FATAL:
		set_prompt(command_giver->ed_buffer->saved_prompt,command_giver);
		if(P_NLINES) clrbuf();

		free((char *)command_giver->ed_buffer);
		command_giver->ed_buffer = 0;
		add_message("FATAL ERROR\n");
		return;
	case CHANGED:
		add_message("File has been changed.\n");
		break;
	case SET_FAIL:
		add_message("`set' command failed.\n");
		break;
	case SUB_FAIL:
		add_message("string substitution failed.\n");
		break;
	case MEM_FAIL:
		add_message("Out of memory: text may have been lost.\n" );
		break;
	default:
		add_message("Unrecognized or failed command.\n");
		/*  Unrecognized or failed command (this  */
		/*  is SOOOO much better than "?" :-)	  */
	}
}

char *query_ed(ob)
   struct object *ob;
{
   if(!ob->ed_buffer) return "\0";
     else if(!ob->ed_buffer->fname || strcmp(ob->ed_buffer->fname,"\0")==0)
             return "New File";
      else return ob->ed_buffer->fname;
}

void clear_buffer(buff)
  struct ed_buffer *buff;
{

	buff->nonascii=0;
	buff->nullchar=0;
	buff->truncated=0;
	buff->oldpat=0;
	buff->CurLn=0;
	buff->LastLn=0;
	buff->Line1=0;
	buff->Line2=0;
	buff->nlines=0;
	buff->nflg=0;
	buff->pflg=0;
	buff->readonly=0;
	buff->nofname=0;
	buff->fchanged=0;
	buff->appending=0;
        buff->block_cmd=0;
}
