#ifndef M_XENIX
#include <stdlib.h>
#endif
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "config.h"
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "realms.h"
#include "error.h"

/*
 * Maintain the wizards high score list about most popular castle.
 */

extern char *string_copy PROT((char *)), *xalloc PROT((int));
extern void add_message();

REALMS *all_realms,*sys_realm;


/*
 * Find the data, if it exists.
 */

int check_realm_access(file_name,ob)
   register char *file_name;
   register OBJECT *ob;
{
	extern struct object_status *ob_status[];
	extern int _check_object PROT((OBJECT *));

   if(!ob_status[0] || !_check_object(ob_status[0]->ob)) return 1;

   if(ob_status[0]->ob->realm==sys_realm) return 1;

   if(ob && _check_object(ob))
     {
       if((ob->realm==sys_realm && ob->flags & GLOBAL) ||
	  (ob->realm==ob_status[0]->ob->realm)) return 1;
       return 0;
     }
   else if(file_name)
    {
      register REALMS *f_realm;

      f_realm=extract_realm(file_name);

      if(f_realm==sys_realm || ob_status[0]->ob->realm==f_realm) return 1;
    }

 return 0;
}

REALMS *extract_realm(path)
  char *path;
{
  register char *p;
  register REALMS *r;

  if(!path) return sys_realm;

  while(path[0]=='/') path++;

  p=strchr(path,'/');

  if(!p) return sys_realm;

  for (r = all_realms; r; r = r->next)
    {
     if(r==sys_realm) continue;
     if(!strncmp(r->realm_dir,path,strlen(r->realm_dir))) return r;
    }

    return sys_realm;
}

REALMS *find_realm(name,flag)
    register char *name;
    int flag;
{
    int length;
    register REALMS *r;

    if(!name) return 0;
    for (r = all_realms; r; r = r->next)
     {
       if (!flag && !strcmp(r->realm_name, name)) return r;
       if(flag==1 && !strcmp(r->realm_owner,name)) return r;
       if(r->realm_dir)
	if(flag==2 && !strcmp(r->realm_dir,name)) return r;
     }
    return 0;
}

/*
 * Check that a name exists. Add it, if it doesn't.
 */

REALMS *add_realm(name,owner,dir)
    char *name,*owner,*dir;
{
    REALMS *r;

    r = (REALMS *)xalloc(sizeof (REALMS));
    r->realm_name = string_copy(name);
    r->realm_owner = string_copy(owner);
    r->realm_dir=string_copy(dir);
    r->fp=0;
    r->next = all_realms;
    all_realms = r;
    if(!strcmp(r->realm_owner,"sysop")) sys_realm=r;
    return r;
}

/*
 * Load the realms file.
 */
void load_realm_file()
{
    FILE *f;
    int i;
    char dir[15],owner[15],name[80];

    f = fopen("info/Realms", "r");
    if(!f && errno == EMFILE) fatal("Max files open.\n");
    if (f == NULL) return;


#if defined(M_XENIX) && !defined(M_UNIX)
    while(fscanf(f,"%[^:]:%[^:]:%[^\n]\n",name,owner,dir) != 0) 
#else
    while(fscanf(f,"%[^:]:%[^:]:%[^\n]\n",name,owner,dir) != EOF) 
#endif
	(void)add_realm(name,owner,dir);

    fclose(f);
}

void realm_list()
{
    REALMS *r, *this_realm;
    int num = 0;
    extern struct object *command_giver;
    extern OBJECT *find_link PROT((OBJECT *));
    int all = 0;

    add_message("\nRealms list\n\n     %-40s %-15s\n\n","Realm","Owner");
    for (r = all_realms; r; r = r->next) {
	if(r==sys_realm) continue;
        num++;
        add_message("%3d. %-40s %-15s\n",num,r->realm_name,r->realm_owner);
    }
    add_message("\n");
}

void remove_realm_list() {
    REALMS *r, *r2;

    for (r = all_realms; r; r = r2) {
	free(r->realm_name);
	free(r->realm_owner);
	free(r->realm_dir);
	r2 = r->next;
	free((char *)r);
    }
}
