#include <stdio.h>
#include <string.h>
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "config.h"
#include "s_hash.h"
#include "realms.h"
#include "error.h"

extern void add_message();

struct s_table_entry {
	char *string;
	struct s_table_entry *next;
	unsigned short ref;
};

extern int hash PROT((char *,int));

static struct s_table_entry *find_s_hash PROT((char *,int));

static struct s_table_entry **s_table;

void init_s_table()
{
   s_table=(struct s_table_entry **)xalloc(sizeof (struct s_table_entry *) * cf.stable);

   memset((char *)s_table,'\0',sizeof (struct s_table_entry *) * cf.stable);
}
static int s_table_count=0;
static int string_bytes=0;
static int strings_malloced=0;

static struct  s_table_entry *add_s_hash(string, hashv)
char			*string;
int			hashv;
{
	extern char *string_copy PROT((char *));
	struct s_table_entry	*tmp;

	tmp=(struct s_table_entry *)xalloc(sizeof (struct s_table_entry));

	tmp->string=string_copy(string);

        string_bytes+=strlen(string);

	strings_malloced+=strlen(string);

	tmp->next = s_table[hashv];

	tmp->ref=1;

	s_table[hashv] = tmp;

	s_table_count++;

	return tmp;
}

static void del_s_hash(entry)
struct s_table_entry	*entry;
{
   struct s_table_entry *tmp, *prev = 0;
   int hashv=hash(entry->string,cf.stable);

   for (tmp = s_table[hashv]; tmp;prev=tmp,tmp = tmp->next)
      if (tmp == entry) {
         if (prev)
            prev->next = tmp->next;
         else
            s_table[hashv] = tmp->next;

	 string_bytes-=strlen(entry->string);
	 strings_malloced-=strlen(entry->string);
	 free(entry->string);
         free((char *)entry);

	 s_table_count--;
         return;
      }
   error(sys_realm,"String did not hash properly\n");

}

char *make_shared_string(string)
register char *string;
{
	register struct s_table_entry *tmp;

        if(tmp=find_s_hash(string,0))
	  {
	     strings_malloced+=strlen(string);
	     tmp->ref++;
	     return tmp->string;
	  }
	tmp=add_s_hash(string,hash(string,cf.stable));

        return tmp->string;
}

void free_string(string,from)
register char *string;
char *from;
{
	register struct s_table_entry *tmp;

	tmp=find_s_hash(string,1);
	if(!tmp)
	  {
		debug_message(sys_realm,"free_shared_string: String '%s' is not in table. from: %s\n",string,from);
	        return;
	  }

	if(--tmp->ref) strings_malloced-=strlen(string);
	   else del_s_hash(tmp);
}

static struct s_table_entry *find_s_hash(string,chk_flag)
register char		*string;
int 			chk_flag;
{
	register struct s_table_entry	*tmp;
	int				hashv;

	hashv = hash(string,cf.stable);

	for(tmp = s_table[hashv]; tmp; tmp = tmp->next) 
	  if((chk_flag && string==tmp->string) || 
	     (!chk_flag && !strcmp(string, tmp->string)))
		  return tmp;
	return 0;
}

void show_s_table_status()
{
  add_message("Strings in hash table               (%5d) : %7d (%9d) bytes\n",cf.stable*4,s_table_count,s_table_count*sizeof (struct s_table_entry));
  add_message("Actual string space        (reduced %5.2f%%) : %7d (%9d) malloced\n",100.00-((float)string_bytes/strings_malloced)*100,string_bytes,strings_malloced);
}
